"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.resolveDependencies = void 0;
const path = require("path");
const resolve = require("enhanced-resolve");
const fs = require("fs");
const source_map_1 = require("source-map");
const transpiler_1 = require("./transpiler");
const utils_1 = require("../utils");
const diagnostics_1 = require("./diagnostics");
const CompilerOptions_1 = require("../CompilerOptions");
const resolver = resolve.ResolverFactory.createResolver({
    extensions: [".lua"],
    enforceExtension: true,
    fileSystem: { ...new resolve.CachedInputFileSystem(fs) },
    useSyncFileSystemCalls: true,
    symlinks: false, // Do not resolve symlinks to their original paths (that breaks node_modules detection)
});
class ResolutionContext {
    constructor(program, options, emitHost) {
        this.program = program;
        this.options = options;
        this.emitHost = emitHost;
        this.diagnostics = [];
        this.resolvedFiles = new Map();
        this.processedDependencies = new Set();
        // value is false if already searched but not found
        this.pathToFile = new Map();
        this.noResolvePaths = new Set(options.noResolvePaths);
    }
    addAndResolveDependencies(file) {
        if (this.resolvedFiles.has(file.fileName))
            return;
        this.resolvedFiles.set(file.fileName, file);
        for (const required of findRequiredPaths(file.code)) {
            // Do not resolve noResolution paths
            if (required.startsWith("@NoResolution:")) {
                // Remove @NoResolution prefix if not building in library mode
                if (!isBuildModeLibrary(this.program)) {
                    const path = required.replace("@NoResolution:", "");
                    replaceRequireInCode(file, required, path);
                    replaceRequireInSourceMap(file, required, path);
                }
                // Skip
                continue;
            }
            // Try to resolve the import starting from the directory `file` is in
            this.resolveImport(file, required);
        }
    }
    resolveImport(file, required) {
        // Do no resolve lualib - always use the lualib of the application entry point, not the lualib from external packages
        if (required === "lualib_bundle") {
            this.resolvedFiles.set("lualib_bundle", { fileName: "lualib_bundle", code: "" });
            return;
        }
        if (this.noResolvePaths.has(required)) {
            if (this.options.tstlVerbose) {
                console.log(`Skipping module resolution of ${required} as it is in the tsconfig noResolvePaths.`);
            }
            return;
        }
        const dependencyPath = this.resolveDependencyPath(file, required);
        if (!dependencyPath)
            return this.couldNotResolveImport(required, file);
        if (this.options.tstlVerbose) {
            console.log(`Resolved ${required} to ${(0, utils_1.normalizeSlashes)(dependencyPath)}`);
        }
        this.processDependency(dependencyPath);
        // Figure out resolved require path and dependency output path
        if (shouldRewriteRequires(dependencyPath, this.program)) {
            const resolvedRequire = (0, transpiler_1.getEmitPathRelativeToOutDir)(dependencyPath, this.program);
            replaceRequireInCode(file, required, resolvedRequire);
            replaceRequireInSourceMap(file, required, resolvedRequire);
        }
    }
    processDependency(dependencyPath) {
        if (this.processedDependencies.has(dependencyPath))
            return;
        this.processedDependencies.add(dependencyPath);
        if (!shouldIncludeDependency(dependencyPath, this.program))
            return;
        // If dependency is not part of project, add dependency to output and resolve its dependencies recursively
        const dependencyContent = this.emitHost.readFile(dependencyPath);
        if (dependencyContent === undefined) {
            this.diagnostics.push((0, diagnostics_1.couldNotReadDependency)(dependencyPath));
            return;
        }
        const dependency = {
            fileName: dependencyPath,
            code: dependencyContent,
        };
        this.addAndResolveDependencies(dependency);
    }
    couldNotResolveImport(required, file) {
        const fallbackRequire = fallbackResolve(required, (0, transpiler_1.getSourceDir)(this.program), path.dirname(file.fileName));
        replaceRequireInCode(file, required, fallbackRequire);
        replaceRequireInSourceMap(file, required, fallbackRequire);
        this.diagnostics.push((0, diagnostics_1.couldNotResolveRequire)(required, path.relative((0, transpiler_1.getProjectRoot)(this.program), file.fileName)));
    }
    resolveDependencyPath(requiringFile, dependency) {
        var _a;
        const fileDirectory = path.dirname(requiringFile.fileName);
        if (this.options.tstlVerbose) {
            console.log(`Resolving "${dependency}" from ${(0, utils_1.normalizeSlashes)(requiringFile.fileName)}`);
        }
        // Check if the import is relative
        const isRelative = ["/", "./", "../"].some(p => dependency.startsWith(p));
        // If the import is relative, always resolve it relative to the requiring file
        // If the import is not relative, resolve it relative to options.baseUrl if it is set
        const relativeTo = isRelative ? fileDirectory : (_a = this.options.baseUrl) !== null && _a !== void 0 ? _a : fileDirectory;
        // Check if file is a file in the project
        const resolvedPath = path.join(relativeTo, dependency);
        const fileFromPath = this.getFileFromPath(resolvedPath);
        if (fileFromPath)
            return fileFromPath;
        // Check if this is a sibling of a required lua file
        if (requiringFile.fileName.endsWith(".lua")) {
            const luaFilePath = resolveLuaPath(fileDirectory, dependency, this.emitHost);
            if (luaFilePath) {
                return luaFilePath;
            }
        }
        // Not a TS file in our project sources, use resolver to check if we can find dependency
        try {
            const resolveResult = resolver.resolveSync({}, fileDirectory, dependency);
            if (resolveResult)
                return resolveResult;
        }
        catch (e) {
            // resolveSync errors if it fails to resolve
        }
        return undefined;
    }
    getFileFromPath(resolvedPath) {
        const existingFile = this.pathToFile.get(resolvedPath);
        if (existingFile)
            return existingFile;
        if (existingFile === false)
            return undefined;
        const file = this.searchForFileFromPath(resolvedPath);
        this.pathToFile.set(resolvedPath, file !== null && file !== void 0 ? file : false);
        return file;
    }
    searchForFileFromPath(resolvedPath) {
        const possibleProjectFiles = [
            resolvedPath,
            resolvedPath + ".ts",
            path.join(resolvedPath, "index.ts"),
            resolvedPath + ".tsx",
            path.join(resolvedPath, "index.tsx"), // tsx index
        ];
        for (const possibleFile of possibleProjectFiles) {
            if (isProjectFile(possibleFile, this.program)) {
                return possibleFile;
            }
        }
        // Check if this is a lua file in the project sources
        const possibleLuaProjectFiles = [
            resolvedPath + ".lua",
            path.join(resolvedPath, "index.lua"),
            path.join(resolvedPath, "init.lua"), // lua looks for <require>/init.lua if it cannot find <require>.lua
        ];
        for (const possibleFile of possibleLuaProjectFiles) {
            if (this.emitHost.fileExists(possibleFile)) {
                return possibleFile;
            }
        }
    }
}
function resolveDependencies(program, files, emitHost) {
    const options = program.getCompilerOptions();
    const resolutionContext = new ResolutionContext(program, options, emitHost);
    // Resolve dependencies for all processed files
    for (const file of files) {
        if (options.tstlVerbose) {
            console.log(`Resolving dependencies for ${(0, utils_1.normalizeSlashes)(file.fileName)}`);
        }
        resolutionContext.addAndResolveDependencies(file);
    }
    return { resolvedFiles: [...resolutionContext.resolvedFiles.values()], diagnostics: resolutionContext.diagnostics };
}
exports.resolveDependencies = resolveDependencies;
function resolveLuaPath(fromPath, dependency, emitHost) {
    const splitDependency = dependency.split(".");
    if (splitDependency.length === 1) {
        // If dependency has just one part (the file), look for a lua file with that name
        const fileDirectory = walkUpFileTreeUntil(fromPath, dir => emitHost.fileExists(path.join(dir, dependency) + ".lua"));
        if (fileDirectory) {
            return path.join(fileDirectory, dependency) + ".lua";
        }
    }
    else {
        // If dependency has multiple parts, look for the first directory of the require path, which must be in the lua root
        const luaRoot = walkUpFileTreeUntil(fromPath, dir => emitHost.directoryExists(path.join(dir, splitDependency[0])));
        if (luaRoot) {
            return path.join(luaRoot, dependency.replace(/\./g, path.sep)) + ".lua";
        }
    }
}
function walkUpFileTreeUntil(fromDirectory, predicate) {
    const currentDir = path.normalize(fromDirectory).split(path.sep);
    while (currentDir.length > 0) {
        const dir = currentDir.join(path.sep);
        if (predicate(dir)) {
            return dir;
        }
        currentDir.pop();
    }
    return undefined;
}
function shouldRewriteRequires(resolvedDependency, program) {
    return !isBuildModeLibrary(program) || !isNodeModulesFile(resolvedDependency);
}
function shouldIncludeDependency(resolvedDependency, program) {
    // Never include lua files (again) that are transpiled from project sources
    if (hasSourceFileInProject(resolvedDependency, program))
        return false;
    // Always include lua files not in node_modules (internal lua sources)
    if (!isNodeModulesFile(resolvedDependency))
        return true;
    // Only include node_modules files if not in library mode
    return !isBuildModeLibrary(program);
}
function isBuildModeLibrary(program) {
    return program.getCompilerOptions().buildMode === CompilerOptions_1.BuildMode.Library;
}
function findRequiredPaths(code) {
    // Find all require("<path>") paths in a lua code string
    const paths = [];
    const pattern = /(^|\s|;|=)require\("(.+?)"\)/g;
    // eslint-disable-next-line @typescript-eslint/ban-types
    let match;
    while ((match = pattern.exec(code))) {
        paths.push(match[2]);
    }
    return paths;
}
function replaceRequireInCode(file, originalRequire, newRequire) {
    const requirePath = (0, utils_1.formatPathToLuaPath)(newRequire.replace(".lua", ""));
    // Escape special characters to prevent the regex from breaking...
    const escapedRequire = originalRequire.replace(/[-/\\^$*+?.()|[\]{}]/g, "\\$&");
    file.code = file.code.replace(new RegExp(`(^|\\s|;|=)require\\("${escapedRequire}"\\)`), `$1require("${requirePath}")`);
}
function replaceRequireInSourceMap(file, originalRequire, newRequire) {
    const requirePath = (0, utils_1.formatPathToLuaPath)(newRequire.replace(".lua", ""));
    if (file.sourceMapNode) {
        replaceInSourceMap(file.sourceMapNode, file.sourceMapNode, `"${originalRequire}"`, `"${requirePath}"`);
    }
}
function replaceInSourceMap(node, parent, require, resolvedRequire) {
    if ((!node.children || node.children.length === 0) && node.toString() === require) {
        parent.children = [new source_map_1.SourceNode(node.line, node.column, node.source, [resolvedRequire])];
        return true; // Stop after finding the first occurrence
    }
    if (node.children) {
        for (const c of node.children) {
            if (replaceInSourceMap(c, node, require, resolvedRequire)) {
                return true; // Occurrence found in one of the children
            }
        }
    }
    return false; // Did not find the require
}
function isNodeModulesFile(filePath) {
    return path
        .normalize(filePath)
        .split(path.sep)
        .some(p => p === "node_modules");
}
function isProjectFile(file, program) {
    return program.getSourceFile(file) !== undefined;
}
function hasSourceFileInProject(filePath, program) {
    const pathWithoutExtension = (0, utils_1.trimExtension)(filePath);
    return (isProjectFile(pathWithoutExtension + ".ts", program) ||
        isProjectFile(pathWithoutExtension + ".tsx", program) ||
        isProjectFile(pathWithoutExtension + ".json", program));
}
// Transform an import path to a lua require that is probably not correct, but can be used as fallback when regular resolution fails
function fallbackResolve(required, sourceRootDir, fileDir) {
    return (0, utils_1.formatPathToLuaPath)(path
        .normalize(path.join(path.relative(sourceRootDir, fileDir), required))
        .split(path.sep)
        .filter(s => s !== "." && s !== "..")
        .join(path.sep));
}
//# sourceMappingURL=resolve.js.map